from typing import Optional, Union, Callable

from . import heuristics
from .active_loop import ActiveLearningLoop
from .dataset import ActiveLearningDataset
from .file_dataset import FileDataset


def get_heuristic(name: str, shuffle_prop: float = 0.0,
                  threshold: Optional[float] = None,
                  reduction: Union[str, Callable] = 'none',
                  **kwargs) -> heuristics.AbstractHeuristic:
    """
    Create an heuristic object from the name.

    Args:
        name (str): Name of the heuristic.
        shuffle_prop (float): Shuffling proportion when getting ranks.
        threshold (Optional[float]): Minimal probability.
        reduction (Union[str, Callable]): Reduction used after computing the score.
        kwargs (dict): Complementary arguments.

    Returns:
        AbstractHeuristic object.
    """
    return {
        'random': heuristics.Random,
        'certainty': heuristics.Certainty,
        'entropy': heuristics.Entropy,
        'margin': heuristics.Margin,
        'bald': heuristics.BALD,
        'variance': heuristics.Variance,
        'precomputed': heuristics.Precomputed,
        'batch_bald': heuristics.BatchBALD
    }[name](shuffle_prop=shuffle_prop, threshold=threshold, reduction=reduction, **kwargs)



import sys
from collections.abc import Sequence
from copy import deepcopy
from typing import Callable, Optional

import numpy as np
import structlog
import torch
from torch.autograd import Variable
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.dataloader import default_collate
from tqdm import tqdm

from baal.utils.iterutils import map_on_tensor
from baal.utils.metrics import Loss

log = structlog.get_logger("ModelWrapper")


class ModelWrapper:
    """
    Wrapper created to ease the training/testing/loading.

    Args:
        model (nn.Module): The model to optimize.
        criterion (Callable): a loss function.
    """

    def __init__(self, model, criterion):
        self.model = model
        self.criterion = criterion
        self.metrics = dict()
        self.add_metric('loss', lambda: Loss())

    def add_metric(self, name: str, initializer: Callable):
        """
        Add a baal.utils.metric.Metric to the Model.

        Args:
            name (str): name of the metric.
            initializer (Callable): lambda to initialize a new instance of a
                                    baal.utils.metrics.Metric object.
        """
        self.metrics['test_' + name] = initializer()
        self.metrics['train_' + name] = initializer()

    def _reset_metrics(self, filter=''):
        """
        Reset all Metrics according to a filter.

        Args:
            filter (str): Only keep the metric if `filter` in the name.
        """
        for k, v in self.metrics.items():
            if filter in k:
                v.reset()

    def _update_metrics(self, out, target, loss, filter=''):
        """
        Update all metrics.

        Args:
            out (Tensor): Prediction
            target (Tensor): Ground truth
            loss (Tensor): Loss from the criterion.
            filter (str): Only update metrics according to this filter.
        """
        for k, v in self.metrics.items():
            if filter in k:
                if 'loss' in k:
                    v.update(loss)
                else:
                    v.update(out, target)

    def train_on_dataset(self, dataset, optimizer, batch_size, epoch, use_cuda, workers=4,
                         collate_fn: Optional[Callable] = None):
        """
        Train for `epoch` epochs on a Dataset `dataset.

        Args:
            dataset (Dataset): Pytorch Dataset to be trained on.
            optimizer (optim.Optimizer): Optimizer to use.
            batch_size (int): The batch size used in the DataLoader.
            epoch (int): Number of epoch to train for.
            use_cuda (bool): Use cuda or not.
            workers (int): Number of workers for the multiprocessing.
            collate_fn (Optional[Callable]): The collate function to use.

        Returns:
            The training history.
        """
        self.train()
        history = []
        log.info("Starting training", epoch=epoch, dataset=len(dataset))
        collate_fn = collate_fn or default_collate
        for _ in range(epoch):
            self._reset_metrics('train')
            for data, target in DataLoader(dataset, batch_size, True, num_workers=workers,
                                           collate_fn=collate_fn):
                _ = self.train_on_batch(data, target, optimizer, use_cuda)
            history.append(self.metrics['train_loss'].value)

        optimizer.zero_grad()  # Assert that the gradient is flushed.
        log.info("Training complete", train_loss=self.metrics['train_loss'].value)
        return history

    def test_on_dataset(
        self,
        dataset: Dataset,
        batch_size: int,
        use_cuda: bool,
        workers: int = 4,
        collate_fn: Optional[Callable] = None,
        average_predictions: int = 1,
    ):
        """
        Test the model on a Dataset `dataset`.

        Args:
            dataset (Dataset): Dataset to evaluate on.
            batch_size (int): Batch size used for evaluation.
            use_cuda (bool): Use Cuda or not.
            workers (int): Number of workers to use.
            collate_fn (Optional[Callable]): The collate function to use.
            average_predictions (int): The number of predictions to average to
                compute the test loss.

        Returns:
            Average loss value over the dataset.
        """
        self.eval()
        log.info("Starting evaluating", dataset=len(dataset))
        self._reset_metrics('test')

        collate_fn = collate_fn or default_collate
        for data, target in DataLoader(dataset, batch_size, False, num_workers=workers,
                                       collate_fn=collate_fn):
            _ = self.test_on_batch(
                data, target, cuda=use_cuda, average_predictions=average_predictions
            )

        log.info("Evaluation complete", test_loss=self.metrics['test_loss'].value)
        return self.metrics['test_loss'].value

    def train_and_test_on_datasets(self, train_dataset: Dataset, test_dataset: Dataset,
                                   optimizer: Optimizer, batch_size: int, epoch: int,
                                   use_cuda: bool,
                                   workers: int = 4,
                                   collate_fn: Optional[Callable] = None,
                                   return_best_weights=False):
        """
        Train and test the model on both Dataset `train_dataset`, `test_dataset`.

        Args:
            train_dataset (Dataset): Dataset to train on.
            test_dataset (Dataset): Dataset to evaluate on.
            optimizer (Optimizer): Optimizer to use during training.
            batch_size (int): Batch size used.
            epoch (int): number of epoch to train on.
            use_cuda (bool): Use Cuda or not.
            workers (int): Number of workers to use.
            collate_fn (Optional[Callable]): The collate function to use.
            return_best_weights (bool): If True, will keep the best weights and return them.

        Returns:
            History and best weights if required.
        """
        best_weight = None
        best_loss = 1e10
        hist = []
        for epoch in range(epoch):
            _ = self.train_on_dataset(train_dataset, optimizer, batch_size, 1,
                                      use_cuda, workers, collate_fn)
            te_loss = self.test_on_dataset(test_dataset, batch_size, use_cuda, workers, collate_fn)
            hist.append({k: v.value for k, v in self.metrics.items()})
            if return_best_weights and te_loss < best_loss:
                best_loss = te_loss
                best_weight = deepcopy(self.state_dict())

        if return_best_weights:
            return hist, best_weight
        else:
            return hist

    def predict_on_dataset_generator(self, dataset: Dataset, batch_size: int, iterations: int,
                                     use_cuda: bool, workers: int = 4,
                                     collate_fn: Optional[Callable] = None, half=False):
        """
        Use the model to predict on a dataset `iterations` time.

        Args:
            dataset (Dataset): Dataset to predict on.
            batch_size (int):  Batch size to use during prediction.
            iterations (int): Number of iterations per sample.
            use_cuda (bool): Use CUDA or not.
            workers (int): Number of workers to use.
            collate_fn (Optional[Callable]): The collate function to use.
            half (bool): if True use half precision.

        Notes:
            The "batch" is made of `batch_size` * `iterations` samples.

        Returns:
            Generators [batch_size, n_classes, ..., n_iterations]
        """
        self.eval()
        if len(dataset) == 0:
            return None

        log.info("Start Predict", dataset=len(dataset))
        collate_fn = collate_fn or default_collate
        loader = DataLoader(dataset,
                            batch_size,
                            False, num_workers=workers,
                            collate_fn=collate_fn)
        for idx, (data, _) in enumerate(tqdm(loader, total=len(loader), file=sys.stdout)):

            pred = self.predict_on_batch(data, iterations, use_cuda)
            pred = map_on_tensor(lambda x: x.detach(), pred)
            if half:
                pred = map_on_tensor(lambda x: x.half(), pred)
            yield map_on_tensor(lambda x: x.cpu().numpy(), pred)

    def predict_on_dataset(self, dataset: Dataset, batch_size: int, iterations: int, use_cuda: bool,
                           workers: int = 4, collate_fn: Optional[Callable] = None, half=False):
        """
        Use the model to predict on a dataset `iterations` time.

        Args:
            dataset (Dataset): Dataset to predict on.
            batch_size (int):  Batch size to use during prediction.
            iterations (int): Number of iterations per sample.
            use_cuda (bool): Use CUDA or not.
            workers (int): Number of workers to use.
            collate_fn (Optional[Callable]): The collate function to use.
            half (bool): if True use half precision

        Notes:
            The "batch" is made of `batch_size` * `iterations` samples.

        Returns:
            Array [n_samples, n_outputs, ..., n_iterations].
        """
        preds = list(self.predict_on_dataset_generator(dataset=dataset, batch_size=batch_size,
                                                       iterations=iterations, use_cuda=use_cuda,
                                                       workers=workers, collate_fn=collate_fn,
                                                       half=half))

        if len(preds) > 0 and not isinstance(preds[0], Sequence):
            # Is an Array or a Tensor
            return np.vstack(preds)
        return [np.vstack(pr) for pr in zip(*preds)]

    def train_on_batch(self, data, target, optimizer, cuda=False):
        """
        Train the current model on a batch using `optimizer`.

        Args:
            data (Tensor): the model input
            target (Tensor): the ground truth
            optimizer (optim.Optimizer): an optimizer
            cuda (bool): use cuda or not

        Returns:
            Tensor, the loss computed from the criterion.
        """
        data = Variable(data)
        if cuda:
            data, target = data.cuda(), target.cuda()
        optimizer.zero_grad()
        output = self.model(data)
        loss = self.criterion(output, target)
        loss.backward()
        optimizer.step()
        self._update_metrics(output, target, loss, filter='train')
        return loss

    def test_on_batch(
        self,
        data: torch.Tensor,
        target: torch.Tensor,
        cuda: bool = False,
        average_predictions: int = 1,
    ):
        """
        Test the current model on a batch.

        Args:
            data (Tensor): the model input
            target (Tensor): the ground truth
            cuda (bool): use cuda or not
            average_predictions (int): The number of predictions to average to
                compute the test loss.

        Returns:
            Tensor, the loss computed from the criterion.
        """
        with torch.no_grad():
            if cuda:
                data, target = data.cuda(), target.cuda()
            if average_predictions == 1:
                preds = self.model(data)
                loss = self.criterion(preds, target)
            elif average_predictions > 1:
                preds = map_on_tensor(lambda p: p.mean(-1),
                                      self.predict_on_batch(data,
                                                            iterations=average_predictions,
                                                            cuda=cuda))
                loss = self.criterion(preds, target)
            self._update_metrics(preds, target, loss, 'test')
            return loss

    def predict_on_batch(self, data, iterations=1, cuda=False):
        """
        Get the model's prediction on a batch.

        Args:
            data (Tensor): the model input
            iterations (int): number of prediction to perform.
            cuda (bool): use cuda or not

        Returns:
            Tensor, the loss computed from the criterion.
                    shape = {batch_size, nclass, n_iteration}
        """
        with torch.no_grad():
            if cuda:
                data = data.cuda()
            input_shape = data.size()
            batch_size = input_shape[0]
            data = torch.stack([data] * iterations)
            data = data.view(batch_size * iterations, *input_shape[1:])
            out = self.model(data)
            out = map_on_tensor(lambda o: o.view([iterations, batch_size, *o.size()[1:]]), out)
            out = map_on_tensor(lambda o: o.permute(1, 2, *range(3, o.ndimension()), 0), out)
            return out

    def get_params(self):
        """
        Return the parameters to optimize.

        Returns:
            Config for parameters.
        """
        return self.model.parameters()

    def state_dict(self):
        """Get the state dict(s)"""
        return self.model.state_dict()

    def load_state_dict(self, state_dict, strict=True):
        """Load the model with `state_dict`"""
        self.model.load_state_dict(state_dict, strict=strict)

    def train(self):
        """Set the model in `train` mode."""
        self.model.train()

    def eval(self):
        """Set the model in `eval mode`."""
        self.model.eval()

    def reset_fcs(self):
        """Reset all torch.nn.Linear layers."""
        def reset(m):
            if isinstance(m, torch.nn.Linear):
                m.reset_parameters()

        self.model.apply(reset)

    def reset_all(self):
        """Reset all *resetable* layers."""
        def reset(m):
            for m in self.model.modules():
                getattr(m, 'reset_parameters', lambda: None)()

        self.model.apply(reset)



from typing import Optional, Callable, Sequence

import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset

from baal import ModelWrapper

available_reductions = {
    'max': lambda x: torch.max(x.view([x.shape[0], -1]), -1, ),
    'min': lambda x: torch.min(x.view([x.shape[0], -1]), -1, ),
    'mean': lambda x: torch.mean(x.view([x.shape[0], -1]), -1, ),
    'sum': lambda x: torch.sum(x.view([x.shape[0], -1]), -1, ),
    'none': lambda x: x,
}


def _shuffle_subset(data: torch.Tensor, shuffle_prop: float) -> torch.Tensor:
    to_shuffle = np.nonzero(np.random.rand(data.shape[0]) < shuffle_prop)[0]
    data[to_shuffle, ...] = data[np.random.permutation(to_shuffle), ...]
    return data


def requireprobs(fn):
    """Will convert logits to probs if needed"""

    def wrapper(self, probabilities):
        # Expected shape : [n_sample, n_classes, ..., n_iterations]
        bounded = torch.min(probabilities) < 0 or torch.max(probabilities) > 1.0
        if bounded or not probabilities.sum(1).allclose(1):
            probabilities = F.softmax(probabilities, 1)
        return fn(self, probabilities)

    return wrapper


class AbstractGPUHeuristic(ModelWrapper):
    """Abstract class that defines a Heuristic.

    Args:
        shuffle_prop (float): shuffle proportion.
        threshold (Optional[float]): threshold the probabilities.
        reverse (bool): True if the most uncertain sample has the highest value.
        reduction (Union[str, Callable]): Reduction used after computing the score.
    """

    def __init__(self, model: ModelWrapper, shuffle_prop=0.0, threshold=None, reverse=False,
                 reduction='none'):
        self.model = model
        self.shuffle_prop = shuffle_prop
        self.threshold = threshold
        self.reversed = reverse
        assert reduction in available_reductions or callable(reduction)
        self.reduction = reduction if callable(reduction) else available_reductions[reduction]

    def compute_score(self, predictions):
        """
        Compute the score according to the heuristic.
        Args:
            predictions (ndarray): Array of predictions

        Returns:
            Array of scores.
        """
        raise NotImplementedError

    def get_uncertainties(self, predictions):
        """Get the uncertainties"""
        scores = self.compute_score(predictions)
        scores = self.reduction(scores)
        scores[~torch.isfinite(scores)] = 0.0 if self.reversed else 10000
        return scores

    def predict_on_batch(self, data, iterations, use_cuda=False):
        """Rank the predictions according to their uncertainties."""
        return self.get_uncertainties(self.model.predict_on_batch(data, iterations, cuda=use_cuda))

    def predict_on_dataset(self, dataset: Dataset, batch_size: int, iterations: int, use_cuda: bool,
                           workers: int = 4, collate_fn: Optional[Callable] = None,
                           half=False):
        """
        Use the model to predict on a dataset `iterations` time.

        Args:
            dataset (Dataset): Dataset to predict on.
            batch_size (int):  Batch size to use during prediction.
            iterations (int): Number of iterations per sample.
            use_cuda (bool): Use CUDA or not.
            workers (int): Number of workers to use.
            collate_fn (Optional[Callable]): The collate function to use.
            half (bool): if True use half precision

        Notes:
            The "batch" is made of `batch_size` * `iterations` samples.

        Returns:
            Array [n_samples, n_outputs, ..., n_iterations].
        """
        preds = list(self.predict_on_dataset_generator(dataset=dataset, batch_size=batch_size,
                                                       iterations=iterations, use_cuda=use_cuda,
                                                       workers=workers, collate_fn=collate_fn,
                                                       half=half))

        if len(preds) > 0 and not isinstance(preds[0], Sequence):
            # Is an Array or a Tensor
            return np.concatenate(preds)
        return [np.concatenate(pr) for pr in zip(*preds)]


class BALDGPUWrapper(AbstractGPUHeuristic):
    """Sort by the highest acquisition function value.
    References:
        https://arxiv.org/abs/1703.02910
    """

    def __init__(self, model: ModelWrapper, shuffle_prop=0.0, threshold=None, reduction='none'):
        super().__init__(model,
                         shuffle_prop=shuffle_prop, threshold=threshold, reverse=True,
                         reduction=reduction
                         )

    @requireprobs
    def compute_score(self, predictions):
        assert predictions.ndimension() >= 3
        # [n_sample, n_class, ..., n_iterations]
        expected_entropy = - torch.mean(torch.sum(predictions * torch.log(predictions + 1e-5), 1),
                                        dim=-1)  # [batch size, ...]
        expected_p = torch.mean(predictions, dim=-1)  # [batch_size, n_classes, ...]
        entropy_expected_p = - torch.sum(expected_p * torch.log(expected_p + 1e-5),
                                         dim=1)  # [batch size, ...]
        bald_acq = entropy_expected_p - expected_entropy
        return bald_acq



from functools import wraps as _wraps
import types
import warnings

import numpy as np
import scipy.stats
from scipy.special import softmax, xlogy
from torch import Tensor

available_reductions = {
    'max': lambda x: np.max(x, axis=tuple(range(1, x.ndim))),
    'min': lambda x: np.min(x, axis=tuple(range(1, x.ndim))),
    'mean': lambda x: np.mean(x, axis=tuple(range(1, x.ndim))),
    'sum': lambda x: np.sum(x, axis=tuple(range(1, x.ndim))),
    'none': lambda x: x,
}


def _shuffle_subset(data: np.ndarray, shuffle_prop: float) -> np.ndarray:
    to_shuffle = np.nonzero(np.random.rand(data.shape[0]) < shuffle_prop)[0]
    data[to_shuffle, ...] = data[np.random.permutation(to_shuffle), ...]
    return data


def singlepass(fn):
    """
    Will take the mean of the iterations if needed.
    Args:
        fn (Callable): Heuristic function.

    Returns:
        fn : Array -> Array

    """

    @_wraps(fn)
    def wrapper(self, probabilities):
        if probabilities.ndim >= 3:
            # Expected shape : [n_sample, n_classes, ..., n_iterations]
            probabilities = probabilities.mean(-1)
        return fn(self, probabilities)

    return wrapper


def requireprobs(fn):
    """Will convert logits to probs if needed"""

    @_wraps(fn)
    def wrapper(self, probabilities):
        # Expected shape : [n_sample, n_classes, ..., n_iterations]
        bounded = np.min(probabilities) < 0 or np.max(probabilities) > 1.0
        if bounded or not np.allclose(probabilities.sum(1), 1):
            probabilities = softmax(probabilities, 1)
        return fn(self, probabilities)

    return wrapper


class AbstractHeuristic:
    """Abstract class that defines a Heuristic.

    Args:
        shuffle_prop (float): shuffle proportion.
        threshold (Optional[float]): threshold the probabilities.
        reverse (bool): True if the most uncertain sample has the highest value.
        reduction (Union[str, Callable]): Reduction used after computing the score.
    """

    def __init__(self, shuffle_prop=0.0, threshold=None, reverse=False, reduction='none'):
        self.shuffle_prop = shuffle_prop
        self.threshold = threshold
        self.reversed = reverse
        assert reduction in available_reductions or callable(reduction)
        self.reduction = reduction if callable(reduction) else available_reductions[reduction]

    def compute_score(self, predictions):
        """
        Compute the score according to the heuristic.
        Args:
            predictions (ndarray): Array of predictions

        Returns:
            Array of scores.
        """
        raise NotImplementedError

    def get_uncertainties_generator(self, predictions):
        """
        Compute the score according to the heuristic.
        Args:
            predictions (Iterable): Generator of predictions

        Returns:
            Array of scores.
        """
        acc = []
        for pred in predictions:
            acc.append(self.get_uncertainties(pred))
        if len(acc) == 0:
            raise ValueError('No prediction! Cannot order the values!')
        return np.concatenate(acc)

    def get_uncertainties(self, predictions):
        """Get the uncertainties"""
        if isinstance(predictions, Tensor):
            predictions = predictions.numpy()
        scores = self.compute_score(predictions)
        scores = self.reduction(scores)
        if not np.all(np.isfinite(scores)):
            fixed = 0.0 if self.reversed else 10000
            warnings.warn(f"Invalid value in the score, will be put to {fixed}", UserWarning)
            scores[~np.isfinite(scores)] = fixed
        return scores

    def get_ranks(self, predictions):
        """Rank the predictions according to their uncertainties."""
        if isinstance(predictions, types.GeneratorType):
            scores = self.get_uncertainties_generator(predictions)
        else:
            scores = self.get_uncertainties(predictions)

        assert scores.ndim <= 2
        ranks = np.argsort(scores, -1)
        if self.threshold:
            ranks = np.asarray([idx for idx in ranks if np.amax(predictions[idx]) > self.threshold])
        if self.reversed:
            ranks = ranks[::-1]
        ranks = _shuffle_subset(ranks, self.shuffle_prop)
        return ranks

    def __call__(self, predictions):
        """Rank the predictions according to their uncertainties."""
        return self.get_ranks(predictions)


class BALD(AbstractHeuristic):
    """Sort by the highest acquisition function value.
    References:
        https://arxiv.org/abs/1703.02910
    """

    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
        super().__init__(
            shuffle_prop=shuffle_prop, threshold=threshold, reverse=True, reduction=reduction
        )

    @requireprobs
    def compute_score(self, predictions):
        assert predictions.ndim >= 3
        # [n_sample, n_class, ..., n_iterations]

        expected_entropy = - np.mean(np.sum(xlogy(predictions, predictions), axis=1),
                                     axis=-1)  # [batch size, ...]
        expected_p = np.mean(predictions, axis=-1)  # [batch_size, n_classes, ...]
        entropy_expected_p = - np.sum(xlogy(expected_p, expected_p),
                                      axis=1)  # [batch size, ...]
        bald_acq = entropy_expected_p - expected_entropy
        return bald_acq


class BatchBALD(BALD):
    """
    Implementation of BatchBALD.

    References:
        https://arxiv.org/abs/1906.08158

    Notes:
        K = iterations, C=classes
        Not tested on 4+ dims.
        """

    def __init__(self, num_samples, shuffle_prop=0.0, threshold=None, reduction='none'):
        self.epsilon = 1e-5
        self.num_samples = num_samples
        super().__init__(shuffle_prop=shuffle_prop, threshold=threshold,
                         reduction=reduction)

    def _conditional_entropy(self, probs):
        K = probs.shape[-1]
        return np.sum(-xlogy(probs, probs), axis=(1, -1)) / K

    def _joint_entropy(self, predictions, selected):
        K = predictions.shape[-1]
        M = selected.shape[0]

        exp_y = np.array(
            [np.matmul(selected, predictions[i].T) for i in range(predictions.shape[0])]) / K
        mean_entropy = selected.mean(-1, keepdims=True)[None]

        step = 256
        for idx in range(0, exp_y.shape[0], step):
            b_preds = exp_y[idx: (idx + step)]
            yield np.sum(-xlogy(b_preds, b_preds) / mean_entropy, axis=(1, -1)) / M

    @requireprobs
    def compute_score(self, predictions):
        MAX_SELECTED = 16000
        MIN_SPREAD = 0.1
        COUNT = 0
        # Get conditional_entropies_B
        conditional_entropies_B = self._conditional_entropy(predictions)
        bald_out = super().compute_score(predictions)
        history = [self.reduction(bald_out).argmax()]
        for step in range(2 * self.num_samples):
            # Select M, iterations example from history, take entropy
            selected = predictions[np.random.permutation(history)[:MAX_SELECTED]]
            selected = self.reduction(scipy.stats.entropy(np.swapaxes(selected, 0, 1)))
            # Compute join entropy
            joint_entropy = list(self._joint_entropy(predictions, selected))
            joint_entropy = np.concatenate(joint_entropy)

            partial_multi_bald_b = joint_entropy - conditional_entropies_B
            partial_multi_bald_b = self.reduction(partial_multi_bald_b)
            partial_multi_bald_b[..., np.array(history)] = -1
            # Add best to history
            partial_multi_bald_b = partial_multi_bald_b.squeeze()
            assert partial_multi_bald_b.ndim == 1
            winner_index = partial_multi_bald_b.argmax()
            history.append(winner_index)
            if partial_multi_bald_b.max() < MIN_SPREAD:
                COUNT += 1
                if COUNT > 50 or len(history) >= predictions.shape[0]:
                    break

        return np.array(history)

    def get_ranks(self, predictions):
        """Rank the predictions according to their uncertainties."""
        if isinstance(predictions, types.GeneratorType):
            raise ValueError("BatchBALD doesn't support generators.")

        ranks = self.get_uncertainties(predictions)

        assert ranks.ndim == 1
        ranks = _shuffle_subset(ranks, self.shuffle_prop)
        return ranks


class Variance(AbstractHeuristic):
    """Sort by the highest variance"""

    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='mean'):
        _help = "Need to reduce the output from [n_sample, n_class] to [n_sample]"
        assert reduction != 'none', _help
        super().__init__(
            shuffle_prop=shuffle_prop, threshold=threshold, reverse=True, reduction=reduction
        )

    def compute_score(self, predictions):
        assert predictions.ndim >= 3
        return np.var(predictions, -1)


class Entropy(AbstractHeuristic):
    """Sort by the highest entropy"""

    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
        super().__init__(
            shuffle_prop=shuffle_prop, threshold=threshold, reverse=True, reduction=reduction
        )

    @singlepass
    @requireprobs
    def compute_score(self, predictions):
        return scipy.stats.entropy(np.swapaxes(predictions, 0, 1))


class Margin(AbstractHeuristic):
    """Sort by the lowest margin, i.e. the difference between the most confident class and
    the second most confident class."""

    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
        super().__init__(
            shuffle_prop=shuffle_prop, threshold=threshold, reverse=False, reduction=reduction
        )

    @singlepass
    @requireprobs
    def compute_score(self, predictions):
        sort_arr = np.sort(predictions, axis=1)
        return sort_arr[:, -1] - sort_arr[:, -2]


class Certainty(AbstractHeuristic):
    """Sort by the lowest certainty"""

    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
        super().__init__(
            shuffle_prop=shuffle_prop, threshold=threshold, reverse=False, reduction=reduction
        )

    @singlepass
    def compute_score(self, predictions):
        return np.max(predictions, axis=1)


class Random(AbstractHeuristic):
    """Random heuristic."""

    def __init__(self, shuffle_prop=0.0, threshold=None, reduction='none'):
        super().__init__(1.0, threshold, False)

    def get_ranks(self, predictions):
        if isinstance(predictions, types.GeneratorType):
            predictions = np.array([1 for _ in predictions])
        ranks = np.arange(predictions.shape[0])
        ranks = _shuffle_subset(ranks, self.shuffle_prop)
        return ranks


class Precomputed(AbstractHeuristic):
    """Precomputed heuristics"""

    def __init__(self, shuffle_prop=0.0, threshold=None, reverse=False):
        super().__init__(shuffle_prop, threshold, reverse=reverse)

    def compute_score(self, predictions):
        return predictions



import types
from typing import Callable

import numpy as np
import torch.utils.data as torchdata

from . import heuristics
from .dataset import ActiveLearningDataset


class ActiveLearningLoop:
    """Object that perform the active learning iteration.

    Args:
        dataset (ActiveLearningDataset): dataset with some sample already labelled.
        get_probabilities (Function): Dataset -> **kwargs ->
                                        ndarray [n_samples, n_outputs, n_iterations]
        heuristic (Heuristic): Heuristic from baal.active.heuristics.
        ndata_to_label (int): Number of sample to label per step.
        max_sample (int): Limit the number of sample used (-1 is no limit).
        **kwargs: parameters forwarded to `get_probabilities`
    """

    def __init__(
        self,
        dataset: ActiveLearningDataset,
        get_probabilities: Callable,
        heuristic: heuristics.AbstractHeuristic = heuristics.Random(),
        ndata_to_label: int = 1,
        max_sample=-1,
        **kwargs,
    ) -> None:
        self.ndata_to_label = ndata_to_label
        self.get_probabilities = get_probabilities
        self.heuristic = heuristic
        self.dataset = dataset
        self.max_sample = max_sample
        self.kwargs = kwargs

    def step(self) -> bool:
        """Perform an active learning step.

        Returns:
            boolean, Flag indicating if we continue training.
        """
        # High to low
        pool = self.dataset.pool

        if len(pool) > 0:

            # Limit number of samples
            if self.max_sample != -1 and self.max_sample < len(pool):
                indices = np.random.choice(len(pool), self.max_sample, replace=False)
                pool = torchdata.Subset(pool, indices)
            else:
                indices = np.arange(len(pool))

            probs = self.get_probabilities(pool, **self.kwargs)
            if probs is not None and (isinstance(probs, types.GeneratorType) or len(probs) > 0):
                to_label = self.heuristic(probs)
                to_label = indices[np.array(to_label)]
                if len(to_label) > 0:
                    self.dataset.label(to_label[: self.ndata_to_label])
                    return True
        return False



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import
from __future__ import print_function

import io
import os
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

from setuptools import find_packages
from setuptools import setup


def read(*names, **kwargs):
    return io.open(join(dirname(__file__), *names), encoding=kwargs.get('encoding', 'utf8')).read()


with open('requirements.txt') as f:
    reqs = [line.strip() for line in f]

with open('documentation-requirements.txt') as f:
    documentation_requirements = [line.strip() for line in f]

setup(
    name='baal',
    version="1.0.0",
    description='Library for bayesian active learning.',
    author='Parmida Atighehchian, Frédéric Branchaud-Charron, Jan Freyberg, Lorne Schell',
    author_email="""parmida@elementai.com,
                    frederic.branchaud-charron@elementai.com,
                    jan.freyberg@elementai.com,
                    lorne.schell@elementai.com""",
    url='https://github.com/ElementAI/baal',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Utilities',
    ],
    install_requires=reqs,
    extras_require={
        'test': ['pytest', 'pytest-pep8', 'hypothesis', 'coverage', 'coveralls'],
        'documentation': documentation_requirements,
    },
)



from copy import copy
from itertools import zip_longest
from typing import Union, Optional, Callable, Tuple, List, Any

import numpy as np
import torch
import torch.utils.data as torchdata


class ActiveLearningDataset(torchdata.Dataset):
    """A dataset that allows for active learning.

    Args:
        dataset (torch.data.Dataset): The baseline dataset.
        eval_transform (Optional(Callable)): transformations to call on the evaluation dataset.
        labelled (Union[np.ndarray, torch.Tensor]):
            An array/tensor that acts as a boolean mask which is True for every
            data point that is labelled, and False for every data point that is not
            labelled.
        make_unlabelled (Callable): the function that returns an
            unlabelled version of a datum so that it can still be used in the DataLoader.
    """

    def __init__(
        self,
        dataset: torchdata.Dataset,
        eval_transform: Optional[Callable] = None,
        labelled: Union[np.ndarray, torch.Tensor] = None,
        make_unlabelled: Callable = lambda x: x,
    ) -> None:
        self._dataset = dataset
        if labelled is not None:
            if isinstance(labelled, torch.Tensor):
                labelled = labelled.numpy()
            self._labelled = labelled.astype(np.bool)
        else:
            self._labelled = np.zeros(len(self._dataset), dtype=np.bool)
        if eval_transform is not None:
            self.eval_transform = eval_transform
        else:
            self.eval_transform = getattr(self._dataset, 'transform', None)

        self.make_unlabelled = make_unlabelled
        # For example, FileDataset has a method 'label'. This is useful when we're in prod.
        self.can_label = hasattr(self._dataset, 'label')

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, ...]:
        """Return stuff from the original dataset."""
        return self._dataset[self._labelled_to_oracle_index(index)]

    def __len__(self) -> int:
        """Return how many actual data / label pairs we have."""
        return self._labelled.sum()

    class ActiveIter():
        """Iterator over an ActiveLearningDataset."""
        def __init__(self, aldataset):
            self.i = 0
            self.aldataset = aldataset

        def __len__(self):
            return len(self.aldataset)

        def __next__(self):
            if self.i >= len(self):
                raise StopIteration

            n = self.aldataset[self.i]
            self.i = self.i + 1
            return n

    def __iter__(self):
        return self.ActiveIter(self)

    @property
    def n_unlabelled(self):
        """The number of unlabelled data points."""
        return (~self._labelled).sum()

    @property
    def n_labelled(self):
        """The number of labelled data points."""
        return self._labelled.sum()

    @property
    def pool(self) -> torchdata.Dataset:
        """Returns a new Dataset made from unlabelled samples"""
        pool_dataset = copy(self._dataset)
        # TODO Handle target transform as well.
        if hasattr(pool_dataset, 'transform') and self.eval_transform is not None:
            pool_dataset.transform = self.eval_transform

        pool_dataset = torchdata.Subset(pool_dataset,
                                        (~self._labelled).nonzero()[0].squeeze())
        ald = ActiveLearningPool(pool_dataset, make_unlabelled=self.make_unlabelled)
        return ald

    """ This returns one or zero, if it is labelled or not, no index is returned.
    """

    def _labelled_to_oracle_index(self, index: int) -> int:
        return self._labelled.nonzero()[0][index].squeeze().item()

    def _pool_to_oracle_index(self, index: Union[int, List[int]]) -> List[int]:
        if isinstance(index, np.int64) or isinstance(index, int):
            index = [index]

        lbl_nz = (~self._labelled).nonzero()[0]
        return [int(lbl_nz[idx].squeeze().item()) for idx in index]

    def _oracle_to_pool_index(self, index: Union[int, List[int]]) -> List[int]:
        if isinstance(index, int):
            index = [index]

        # Pool indices are the unlabelled, starts at 0
        lbl_cs = np.cumsum(~self._labelled) - 1
        return [int(lbl_cs[idx].squeeze().item()) for idx in index]

    def label(self, index: Union[list, int], value: Optional[Any] = None) -> None:
        """
        Label data points.
        The index should be relative to the pool, not the overall data.

        Args:
            index (Union[list,int]): one or many indices to label.
            value (Optional[Any]): The label value. If not provided, no modification
                                    to the underlying dataset is done.
        """
        if isinstance(index, int):
            index = [index]
        if not isinstance(value, (list, tuple)):
            value = [value]
        indexes = self._pool_to_oracle_index(index)
        for index, val in zip_longest(indexes, value, fillvalue=None):
            if self.can_label and val is not None:
                self._dataset.label(index, val)
            self._labelled[index] = 1

    def label_randomly(self, n: int = 1) -> None:
        """
        Label `n` data-points randomly.

        Args:
            n (int): number of samples to label.
        """
        for i in range(n):
            """Making multiple call to self.n_unlabelled is eneficient, but
            self.label changes the available length and it may lead to
            IndexError if not done this way."""
            self.label(np.random.choice(self.n_unlabelled, 1).item())

    def reset_labeled(self):
        """Reset the label map."""
        self._labelled = np.zeros(len(self._dataset), dtype=np.bool)

    def is_labelled(self, idx: int) -> bool:
        """Check if a datapoint is labelled"""
        return self._labelled[idx] == 1

    def get_raw(self, idx: int) -> None:
        """Get a datapoint from the underlying dataset."""
        return self._dataset[idx]

    def state_dict(self):
        """Return the state_dict, ie. the labelled map."""
        return {'labeled': self._labelled}


class ActiveLearningPool(torchdata.Dataset):
    """ A dataset that represents the unlabelled pool for active learning.

    Args:
        dataset (Dataset): A Dataset object providing unlabeled sample.
        make_unlabelled (Callable): the function that returns an
            unlabelled version of a datum so that it can still be used in the DataLoader.

    """

    def __init__(self, dataset: torchdata.Dataset, make_unlabelled: Callable = lambda x: x) -> None:
        self._dataset = dataset
        self.make_unlabelled = make_unlabelled

    def __getitem__(self, index: int) -> Tuple[torch.Tensor, ...]:
        # this datum is marked as unlabelled, so clear the label
        return self.make_unlabelled(self._dataset[index])

    def __len__(self) -> int:
        """Return how many actual data / label pairs we have."""
        return len(self._dataset)
